#include <algorithm>
#include "GLSL.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "Camera.h"

#define PI 3.1415926f
#define DEG_TO_RAD PI / 180.0f
#define RAD_TO_DEG 180.0f / PI
#define toDeg(x) ((x) * RAD_TO_DEG)
#define toRad(x) ((x) * DEG_TO_RAD)

// Initialize a Camera
Camera::Camera() {
	eye = vec3(0, EYE_START, 0);
	curX = curY = lastY = phi = theta = 0;
}

void Camera::Init(GLFWwindow *window, float g_width, float g_height) {
	glfwGetCursorPos(window, &curX, &lastY);
	curY = g_height/2;
	cursor_position_callback(window, curX, lastY, g_width, g_height);
}

// Modify |eye| position based on polled key 
// Translations take the magnitude |speed| in the specified direction as applied over |elapsed| seconds
void Camera::translate(GLFWwindow *window, vec3 look, float elapsed) {
	vec3 movedir(0, 0, 0);
	float speed = SPEED * elapsed;

	look = glm::normalize(look);

	if (glfwGetKey(window, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS) // Speed up with SHIFT
		speed *= SPEED_FAST;

	if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS) // Move along toward viewpoint
		movedir += look;
	if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS) // Move away from viewpoint
		movedir -= look;

	if (glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS) // Strafe left
		movedir += glm::normalize(glm::cross(UP_VECTOR, look));
	if (glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS) // Strafe right
		movedir -= glm::normalize(glm::cross(UP_VECTOR, look));
	
	if (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS) // Climb upward
		movedir += UP_VECTOR;
	if (glfwGetKey(window, GLFW_KEY_LEFT_CONTROL) == GLFW_PRESS) // Sink downward
		movedir -= UP_VECTOR;

	if (movedir.x || movedir.y || movedir.z) {
		eye += speed * glm::normalize(movedir); // Constant speed along the vector specified
		clamp();
	}
	
	//fixForScreenshot();
}

// Ensure the camera doesn't go out of bounds
void Camera::clamp() {
	if (eye.y < EYE_MIN + EYE_SIZE)
		eye.y = EYE_MIN + EYE_SIZE;
	else if (eye.y > EYE_MAX - EYE_SIZE)
		eye.y = EYE_MAX - EYE_SIZE;
}

void Camera::cursor_position_callback(GLFWwindow* window, double xpos, double ypos, float g_height, float g_width)
{
	curX = xpos;
	curY = std::max(std::min(curY + ypos - lastY, (double)g_height), 0.0);
	lastY = ypos;
	phi = toRad(MAX_PITCH) - toRad(2 * MAX_PITCH) * curY / g_height;
	theta = -toRad(MAX_PITCH) + toRad(2 * MAX_PITCH) * curX / g_width;

	//fixForScreenshot();
	
	lookat = glm::normalize(vec3(cos(phi)*cos(theta), sin(phi), cos(phi)*cos(PI / 2 - theta)));
}

mat4 Camera::getView() {
	return glm::lookAt(eye, eye + lookat, UP_VECTOR);
}

mat4 Camera::getViewLookat(vec3 look) {
	return glm::lookAt(eye, look, UP_VECTOR);
}

void Camera::fixForScreenshot() {
	eye = vec3(24.1f, 6.74f, 69.54f);
	phi = 0.1473f;
	theta = -4.084;
}